from __future__ import annotations

from typing import (Iterable, Mapping, Sequence, Type, TypeVar, Union, Any,
                    Tuple)

from .expressions import ExprLike
from .fields import Field
from .protocols import Adapter
from .queries.delete_all import DeleteAllQuery
from .queries.delete_object import DeleteObjectQuery
from .queries.insert_all import InsertAllQuery
from .queries.insert_object import InsertObjectQuery
from .queries.raw import RawQuery
from .queries.select import SelectQuery
from .queries.update_all import UpdateAllQuery
from .queries.update_object import UpdateObjectQuery
from .query import QueryBuilder
from .schemas import Schema

E = TypeVar("E", bound=Schema)


class Store:
    def __init__(self, adapter: Adapter):
        self._adapter = adapter

    def raw(self, query: str, params: Mapping = None) -> RawQuery[E]:
        return RawQuery(
            adapter=self._adapter,
            query=query,
            values=params,
        )

    def query(self, query: Union[QueryBuilder, Type[E]]) -> SelectQuery[E]:
        if issubclass(query, Schema):
            query = QueryBuilder(query)
        return SelectQuery(adapter=self._adapter, query=query)

    async def get(self, schema: Type[E], pk: Union[Any, Tuple[Any]]) -> E:
        query = QueryBuilder(schema).where(schema.pk == pk)
        return await self.query(query).one()

    async def get_by(self, schema: Type[E], where: ExprLike) -> E:
        query = QueryBuilder(schema).where(where)
        return await self.query(query).one_or_none()

    def insert(self, entity: E) -> InsertObjectQuery[E]:
        return InsertObjectQuery(
            adapter=self._adapter,
            entity=entity,
        )

    def insert_all(
            self, schema: Type[E], values: Iterable[Mapping],
            batch_size: int = None,
    ) -> InsertAllQuery:
        return InsertAllQuery(
            adapter=self._adapter,
            schema=schema,
            values=values,
            batch_size=batch_size,
        )

    def update(self, entity: E, **values) -> UpdateObjectQuery[E]:
        return UpdateObjectQuery(
            adapter=self._adapter,
            entity=entity,
            values=values,
        )

    def update_all(
            self,
            schema: Type[E],
            values: Mapping,
            where: ExprLike = None,
            returning: Union[Field, Sequence[Field]] = None,
    ) -> UpdateAllQuery:
        return UpdateAllQuery(
            adapter=self._adapter,
            schema=schema,
            values=values,
            where=where,
            returning=returning,
        )

    def delete(self, entity: E) -> DeleteObjectQuery:
        return DeleteObjectQuery(
            adapter=self._adapter,
            entity=entity,
        )

    def delete_all(
            self, schema: Type[E],
            where: ExprLike = None,
    ) -> DeleteAllQuery:
        return DeleteAllQuery(
            adapter=self._adapter,
            schema=schema,
            where=where,
        )



from typing import (Generic, Mapping, Sequence, Type, TypeVar, Union)

from aerie.expressions import ExprLike
from aerie.fields import Field
from aerie.protocols import Adapter
from aerie.queries.base import AwaitableQuery
from aerie.schemas import Schema

T = TypeVar('T', bound=Schema)


class UpdateAllQuery(Generic[T], AwaitableQuery):
    def __init__(
            self,
            adapter: Adapter,
            schema: Type[T],
            values: Mapping,
            where: ExprLike = None,
            returning: Union[Field, Sequence[Field]] = None,
    ):
        self._schema = schema
        self._adapter = adapter
        self._where = where
        self._values = values
        self._returning = returning

    async def execute(self) -> Sequence[Mapping]:
        return await self._adapter.update_all(
            schema=self._schema,
            values=self._values,
            where=self._where,
            returning=self._returning,
        )



from __future__ import annotations

import datetime
from typing import (AbstractSet, Any, Dict, Iterator, List, Mapping, Optional,
                    Sequence, TYPE_CHECKING, Tuple, Type, Union)

import sqlalchemy as sa

from .expressions import Conditions, Expr, Nulls, OrderBy, Sort, and_
from .mappers import object_to_dict
from .utils import make_table_name

if TYPE_CHECKING:
    from .protocols import Queryable
    from .schemas import Schema


class Sorting:
    def asc(
            self: Field, nulls: Nulls = Nulls.UNDEFINED
    ) -> OrderBy:
        return OrderBy(self, Sort.ASC, nulls)

    def desc(
            self: Field,
            nulls: Nulls = Nulls.UNDEFINED
    ) -> OrderBy:
        return OrderBy(self, Sort.DESC, nulls)


class Lookups:

    def concat(self: Field, other) -> Expr:
        return Expr(self, Conditions.CONCAT, other)

    def like(self: Field, other) -> Expr:
        return Expr(self, Conditions.LIKE, other)

    def not_like(self: Field, other) -> Expr:
        return ~self.like(other)

    def ilike(self: Field, other) -> Expr:
        return Expr(self, Conditions.ILIKE, other)

    def not_ilike(self: Field, other) -> Expr:
        return ~self.ilike(other)

    def in_(self: Field, other: Union[Sequence, Queryable]) -> Expr:
        return Expr(self, Conditions.IN, other)

    def not_in(self: Field, other: Union[Sequence, Queryable]) -> Expr:
        return ~self.in_(other)

    def is_(self: Field, other) -> Expr:
        return Expr(self, Conditions.IS, other)

    def startswith(self: Field, other) -> Expr:
        return Expr(self, Conditions.STARTS_WITH, other)

    def not_startswith(self: Field, other) -> Expr:
        return ~self.startswith(other)

    def endswith(self: Field, other) -> Expr:
        return Expr(self, Conditions.ENDS_WITH, other)

    def not_endswith(self: Field, other) -> Expr:
        return ~self.endswith(other)

    def contains(self: Field, other) -> Expr:
        return Expr(self, Conditions.CONTAINS, other)

    def not_contains(self: Field, other) -> Expr:
        return ~self.contains(other)

    def match(self: Field, other) -> Expr:
        return Expr(self, Conditions.MATCH, other)

    def not_match(self: Field, other) -> Expr:
        return ~self.match(other)

    def between(self: Field, left, right) -> Expr:
        return Expr(self, Conditions.BETWEEN, (left, right))

    def not_between(self: Field, left, right) -> Expr:
        return ~self.between(left, right)

    def __eq__(self: Field, other) -> Expr:
        return Expr(self, Conditions.EQUALS, other)

    def __ne__(self: Field, other) -> Expr:
        return ~self.__eq__(other)

    def __gt__(self: Field, other) -> Expr:
        return Expr(self, Conditions.GREATER, other)

    def __ge__(self: Field, other) -> Expr:
        return Expr(self, Conditions.GREATER_OR_EQUAL, other)

    def __lt__(self: Field, other) -> Expr:
        return Expr(self, Conditions.LESSER, other)

    def __le__(self: Field, other) -> Expr:
        return Expr(self, Conditions.LESSER_OR_EQUAL, other)

    def __neg__(self: Field) -> Expr:
        return Expr(self, Conditions.NEGATE)

    def __add__(self: Field, other) -> Expr:
        return Expr(self, Conditions.ADD, other)

    def __radd__(self: Field, other) -> Expr:
        return Expr(self, Conditions.RIGHT_ADD, other)

    def __sub__(self: Field, other) -> Expr:
        return Expr(self, Conditions.SUBTRACT, other)

    def __rsub__(self: Field, other) -> Expr:
        return Expr(self, Conditions.RIGHT_SUBTRACT, other)

    def __mul__(self: Field, other) -> Expr:
        return Expr(self, Conditions.MULTIPLY, other)

    def __rmul__(self: Field, other) -> Expr:
        return Expr(self, Conditions.RIGHT_MULTIPLY, other)

    def __truediv__(self: Field, other) -> Expr:
        return Expr(self, Conditions.DIVIDE, other)

    def __rtruediv__(self: Field, other) -> Expr:
        return Expr(self, Conditions.RIGHT_DIVIDE, other)

    def __mod__(self: Field, other) -> Expr:
        return Expr(self, Conditions.MODULO, other)

    def __rmod__(self: Field, other) -> Expr:
        return Expr(self, Conditions.RIGHT_MODULO, other)


class Field(Lookups, Sorting):
    type = None
    _creation_counter = 0

    def __new__(cls, *args, **kwargs):
        cls._creation_counter += 1
        instance = super().__new__(cls)
        instance._creation_counter = cls._creation_counter
        instance.__init__(*args, **kwargs)
        return instance

    def __init__(
            self,
            *,
            attr_name: str = None,
            default: Any = None,
            null: bool = False,
            source_name: str = None,
            primary_key: bool = False,
            schema: Type[Schema] = None,
    ):
        self.attr_name = attr_name
        self.null = null
        self.default = default
        self.source_name = source_name or attr_name
        self.primary_key = primary_key
        self.schema = schema

    def to_db_value(self, value: Any) -> Any:
        """Convert python object to database value."""
        if value is None or type(value) == self.type:
            return value
        return self.type(value)

    def to_python_value(self, value: Any) -> Any:
        """Convert database value to python object."""
        if value is None or isinstance(value, self.type):
            return value
        return self.type(value)

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self}>"

    def __str__(self):
        return self.attr_name or '<not bound>'

    def __get__(self, instance: Field, owner: Type[Field]):
        self.schema = owner
        if instance is None:
            return self
        return instance.__dict__.get(self.attr_name)

    def __set__(self, instance: Field, value: Any):
        instance.__dict__[self.attr_name] = value

    def __hash__(self):
        return self._creation_counter


class Integer(Field):
    type = int

    def __init__(self, *, auto_increment: bool = False, **kwargs):
        super().__init__(**kwargs)
        self.auto_increment = auto_increment


class FloatField(Field):
    type = float


class String(Field):
    type = str


class TextField(Field):
    type = str


class BooleanField(Field):
    type = bool

    def to_python_value(self, value):
        if isinstance(value, bool):
            return value
        return str(value).strip().lower() in ["yes", "1", "on", "true"]


class DateField(Field):
    type = datetime.date

    def to_db_value(self, value: datetime.date) -> Optional[str]:
        if value is None:
            return value
        return value.isoformat()

    def to_python_value(self,
                        value: Union[str, datetime.date]) -> datetime.date:
        if value is None or isinstance(value, self.type):
            return value

        return datetime.date.fromisoformat(value)


class DateTimeField(Field):
    type = datetime.datetime

    def to_db_value(self, value: datetime.datetime) -> Optional[str]:
        if value is None:
            return value
        return value.isoformat()

    def to_python_value(
            self, value: Union[str, datetime.datetime]
    ) -> datetime.datetime:
        if value is None or isinstance(value, self.type):
            return value
        return datetime.datetime.fromisoformat(value)


class TimeField(Field):
    type = datetime.time

    def to_db_value(self, value: datetime.time) -> Optional[str]:
        if value is None:
            return value
        return value.isoformat()

    def to_python_value(self,
                        value: Union[str, datetime.time]) -> datetime.time:
        if value is None or isinstance(value, self.type):
            return value

        return datetime.time.fromisoformat(value)


class JsonField(Field):
    type = dict


class EmbedsOne(Field):
    type = dict

    def __init__(self, schema: Type["Schema"], *args, **kwargs):
        super().__init__(*args, schema=schema, **kwargs)

    def to_db_value(self, value):
        if value is None:
            return value
        return object_to_dict(value)

    def to_python_value(self, value):
        if value is None:
            return value
        return self.schema(**value)


class EmbedsMany(Field):
    type = list

    def __init__(self, schema: Type["Schema"], *args, **kwargs):
        super().__init__(*args, schema=schema, **kwargs)

    def to_db_value(self, value):
        if value is None:
            return value
        return [object_to_dict(item) for item in value]

    def to_python_value(self, value):
        if value is None:
            return value
        return [self.schema(**item) for item in value]


class PrimaryKey:
    def __init__(self, *fields: Field):
        self.fields = fields
        self._names = [f.attr_name for f in fields]

    @property
    def is_composite(self) -> bool:
        return len(self.fields) > 1

    def asc(self) -> List[OrderBy]:
        return [f.asc().nulls_last() for f in self.fields]

    def desc(self) -> List[OrderBy]:
        return [f.desc().nulls_last() for f in self.fields]

    def __contains__(self, field: Union[str, Field]) -> bool:
        if isinstance(field, str):
            return field in self._names

        return field.attr_name in self._names

    # todo: needs test
    def __eq__(self, other):
        if isinstance(other, (str, int, float)):
            other = [other]

        return and_(*[f == value for f, value in zip(self.fields, other)])

    # todo: needs test
    def __ne__(self, other):
        if isinstance(other, (str,)):
            other = [other]
        return sa.and_(*[f != other[i] for i, f in enumerate(self.fields)])

    def __getitem__(self, item):
        if item > len(self.fields) - 1:
            raise IndexError()
        return self.fields[item]

    def __iter__(self) -> Iterator[Field]:
        return iter(self.fields)

    def __next__(self):
        return next(self.__iter__())

    def __len__(self):
        return len(self.fields)

    def __get__(
            self, instance: Schema, owner: Type[Schema]
    ) -> Union[PrimaryKey, Tuple]:
        if instance is None:
            return self

        return tuple(
            getattr(instance, pk_field.attr_name)
            for pk_field in self.fields
        )

    def __set__(self, instance: Schema, value):
        """Set primary key value to entity instance."""
        if len(self) == 1 and not isinstance(value, Sequence):
            value = [value]

        for index, field in enumerate(self):
            setattr(instance, field.attr_name, value[index])

    def __repr__(self):
        return f'<PrimaryKey: {[f.attr_name for f in self.fields]}>'


class Relation:
    schema: Type[Schema]
    attr_name: str = None
    source_name: str = None

    def alter_schema_fields(self, fields: Dict[str, Field]):
        pass


undefined = object()


class HasOne(Relation):
    def __init__(
            self,
            schema: Type[Schema],
            attr_name: str = None,
            source_name: str = None,
            foreign_key: str = 'id',
            null: bool = True,
            default: Any = None,
            field: Field = None,
    ):
        if source_name is None:
            source_name = make_table_name(schema.__name__) + '_id'

        if field is None:
            field = Integer(schema=schema)

        if foreign_key is None:
            foreign_key = 'id'

        if not null and not default:
            raise ValueError(
                'Non-nullable relations require "default" attribute set.'
            )

        self.schema = schema
        self.attr_name = attr_name
        self.foreign_key = foreign_key
        self.source_name = source_name
        self.field = field
        self.null = null
        self.default = default
        self.value = schema.AssociationNotLoaded(
            'Access to not preloaded association. '
            'You must explicitly preload associations before accessing them. '
        )

    def alter_schema_fields(self, fields: Dict[str, Field]):
        if self.source_name not in fields:
            fields[self.source_name] = self.field

    def __get__(self, instance: Schema, owner: Type[Schema]):
        if instance is None:
            return self
        return self.value


class Fields:
    def __init__(self, items: Mapping[str, Field]):
        self._items = items

    @property
    def _by_source(self):
        return {f.source_name: f for f in self._items.values()}

    def values(self) -> List[Field]:
        return list(self._items.values())

    def items(self) -> AbstractSet[Tuple[str, Field]]:
        return self._items.items()

    def by_source_name(self, name: str) -> Field:
        return self._by_source[name]

    def __iter__(self):
        return iter(self._items.values())

    def __getitem__(self, item: str) -> Field:
        return self._items[item]

    def __setitem__(self, name: str, value: Field):
        self._items[name] = value

    def __contains__(self, item: Union[str, Field]) -> bool:
        if isinstance(item, Field):
            item = item.attr_name
        return item in self._items



from typing import Generic, Type, TypeVar

from aerie.expressions import ExprLike
from aerie.protocols import Adapter
from aerie.queries.base import AwaitableQuery
from aerie.schemas import Schema

T = TypeVar('T', bound=Schema)


class DeleteAllQuery(Generic[T], AwaitableQuery):
    def __init__(
            self,
            adapter: Adapter,
            schema: Type[T],
            where: ExprLike = None,
    ):
        self._adapter = adapter
        self._schema = schema
        self._where = where

    async def execute(self) -> None:
        await self._adapter.delete_all(
            schema=self._schema,
            where=self._where,
        )



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from aerie.schemas import Schema


class AerieException(Exception):
    ...


class DoesNotExist(AerieException):
    ...


class MultipleResults(AerieException):
    ...


class ImproperlyConfigured(AerieException):
    ...


class AssociationNotLoaded(AttributeError, AerieException):
    def _raise(self):
        raise self

    def __get__(self, instance, owner):
        self._raise()

    def __getattr__(self, item):
        self._raise()

    def __str__(self):
        return ''


class FieldError(AttributeError, AerieException):
    pass


class UniqueViolation(AerieException):
    def __init__(
            self, msg: str, hint: str, code: str, constraint: str,
            query: str,
    ):
        super().__init__(msg)
        self.hint = hint
        self.code = code
        self.query = query
        self.constraint = constraint


class DeletedEntity(AerieException):
    def __init__(self, msg: str, entity: Schema):
        super().__init__(msg)
        self.entity = entity



from typing import Type

import sqlalchemy as sa

from aerie import fields
from aerie.expressions import Conditions, ExprLike, LogicalExpr, Expr
from aerie.fields import Field, HasOne
from aerie.schemas import Schema

# a map of aerie fields to sqlalchemy column types
COLUMNS_MAP = {
    fields.Integer: sa.Integer,
    fields.String: sa.Text,
    fields.BooleanField: sa.Boolean,
    fields.DateField: sa.Date,
    fields.DateTimeField: sa.DateTime,
    fields.TimeField: sa.Time,
    fields.JsonField: sa.JSON,
    fields.EmbedsOne: sa.JSON,
    fields.EmbedsMany: sa.JSON,
}

# a map of converters from aerie expression to sqlalchemy clause
CONVERTERS_MAP = {
    Conditions.CONCAT: lambda col, value: col.concat(value),
    Conditions.LIKE: lambda col, value: col.like(value),
    Conditions.ILIKE: lambda col, value: col.ilike(value),
    Conditions.IN: lambda col, value: col.in_(value),
    Conditions.IS: lambda col, value: col.is_(value),
    Conditions.STARTS_WITH: lambda col, value: col.startswith(value),
    Conditions.ENDS_WITH: lambda col, value: col.endswith(value),
    Conditions.CONTAINS: lambda col, value: col.contains(value),
    Conditions.MATCH: lambda col, value: col.match(value),
    Conditions.BETWEEN: lambda col, value: col.between(*value),
    Conditions.EQUALS: lambda col, value: col == value,
    Conditions.GREATER: lambda col, value: col > value,
    Conditions.GREATER_OR_EQUAL: lambda col, value: col >= value,
    Conditions.LESSER: lambda col, value: col < value,
    Conditions.LESSER_OR_EQUAL: lambda col, value: col <= value,
    Conditions.NEGATE: lambda col, value: ~col,
    Conditions.ADD: lambda col, value: col + value,
    Conditions.RIGHT_ADD: lambda col, value: value + col,
    Conditions.SUBTRACT: lambda col, value: col - value,
    Conditions.RIGHT_SUBTRACT: lambda col, value: value - col,
    Conditions.MULTIPLY: lambda col, value: col * value,
    Conditions.RIGHT_MULTIPLY: lambda col, value: value * col,
    Conditions.DIVIDE: lambda col, value: col / value,
    Conditions.RIGHT_DIVIDE: lambda col, value: value / col,
    Conditions.MODULO: lambda col, value: col % value,
    Conditions.RIGHT_MODULO: lambda col, value: value % col,
}

CONDITIONS_MAP = {
    Conditions.AND: sa.sql.and_,
    Conditions.OR: sa.sql.or_,
}


def generate_table(
        schema: Type[Schema],
        metadata: sa.MetaData,
) -> sa.Table:
    source = schema.__meta__.source
    fields_ = schema.__meta__.fields.values()

    if source in metadata.tables:
        return metadata.tables[source]

    related_columns = []
    for relation in schema.__meta__.relations.values():
        if isinstance(relation, HasOne):
            fk_table = generate_table(relation.schema, metadata)
            fk_column = fk_table.c[relation.foreign_key]
            related_columns.append(sa.Column(
                relation.source_name,
                COLUMNS_MAP[relation.field.__class__],
                sa.ForeignKey(fk_column),
                nullable=relation.null,
                default=relation.default,
            ))

    return sa.Table(
        source, metadata,
        *[
            sa.Column(
                field.source_name,
                COLUMNS_MAP[field.__class__],
                nullable=field.null,
                default=field.default,
                autoincrement=getattr(field, 'auto_increment', None),
                primary_key=field in schema.__meta__.primary_key,
            ) for field in fields_
        ],
        *related_columns
    )


def convert_expression(
        expr: ExprLike,
        metadata: sa.MetaData,
) -> sa.sql.ClauseElement:
    if isinstance(expr, LogicalExpr):
        if expr.condition not in CONDITIONS_MAP:
            raise KeyError(
                f'Condition "{expr.condition}" is not supported by SQL adapter.'
            )

        return CONDITIONS_MAP[expr.condition](
            convert_expression(expr.left, metadata),
            convert_expression(expr.right, metadata),
        )
    else:
        table = generate_table(expr.field.schema, metadata)
        column = table.c[expr.field.source_name]
        value = expr.value
        if isinstance(value, Field):
            value_table = generate_table(expr.value.schema, metadata)
            value = value_table.c[value.source_name]
        if isinstance(value, (Expr, LogicalExpr)):
            value = convert_expression(value, metadata)
        built = CONVERTERS_MAP[expr.condition](column, value)
        return ~built if expr.negate else built



from typing import Dict, Generic, Iterable, Mapping, Type, TypeVar, Union

from aerie.collections import Collection
from aerie.protocols import Adapter
from aerie.queries.base import AwaitableQuery
from aerie.schemas import Schema

T = TypeVar('T', bound=Schema)


class InsertAllQuery(Generic[T], AwaitableQuery):
    def __init__(
            self,
            adapter: Adapter,
            schema: Type[T],
            values: Union[Mapping, Iterable[Mapping]],
            batch_size: int = None,
    ):
        self._schema = schema
        self._adapter = adapter
        self._batch_size = batch_size
        if isinstance(values, Dict):
            values = [values]
        self._values = Collection(values)

    async def execute(self) -> None:
        await self._adapter.insert_all(
            self._schema,
            self._values,
            self._batch_size,
        )



from typing import Generic, TypeVar

from aerie.protocols import Adapter
from aerie.queries.base import AwaitableQuery
from aerie.schemas import Schema, State

T = TypeVar('T', bound=Schema)


class DeleteObjectQuery(Generic[T], AwaitableQuery):
    def __init__(self, adapter: Adapter, entity: T):
        self._adapter = adapter
        self._entity = entity

    async def execute(self) -> T:
        schema = type(self._entity)

        await self._adapter.delete_all(
            schema=schema,
            where=schema.pk == self._entity.pk,
        )
        self._entity.__meta__.state = State.DELETED
        return self._entity



from __future__ import annotations
from typing import Dict, Mapping, Type, TypeVar, TYPE_CHECKING

if TYPE_CHECKING:
    from .schemas import Schema

E = TypeVar("E", bound='Schema')


def entity_constructor(row: Mapping, entity: Type[E]) -> E:
    if hasattr(entity, "__meta__"):
        row = {
            k: f.to_python_value(row.get(k, None))
            for k, f in entity.__meta__.fields.items()
        }

    return entity(**row)


def object_to_dict(entity: E) -> Dict:
    if hasattr(entity, "__meta__"):
        mapping = {}
        for attr_name, field in type(entity).__meta__.fields.items():
            mapping[field.source_name] = field.to_db_value(
                getattr(entity, attr_name)
            )

        for attr_name, relation in type(entity).__meta__.relations.items():
            if not relation.many:
                related_object = getattr(entity, attr_name)
                if not isinstance(
                        related_object, relation.schema.AssociationNotLoaded
                ):
                    mapping[relation.source_name] = relation.to_db_value(
                        related_object
                    )
        return mapping

    return {
        k: v for k, v in entity.__dict__
        if not k.startswith("_") and not k.isupper()
    }



from __future__ import annotations

import enum
from typing import (AsyncGenerator, List, Literal, Mapping, Protocol, Sequence,
                    TYPE_CHECKING, Type, TypeVar, Union, Any)

from .expressions import ExprLike
from .query import QueryBuilder
from .schemas import Schema

if TYPE_CHECKING:
    from .schemas import Metadata, Schema
    from .fields import Field


class SchemaLike(Protocol):
    __meta__: Metadata


class OnConflict(enum.Enum):
    IGNORE = 'ignore'
    UPDATE = 'update'


E = TypeVar('E', bound=Schema)
MAP_TO = TypeVar('MAP_TO')


class Queryable(Protocol):
    def to_expr(self):
        raise NotImplementedError()


class Adapter(Protocol[E]):

    def insert_all(
            self,
            schema: Type[E],
            values: Sequence[Mapping],
            batch_size: int = None,
            on_conflict: Literal[OnConflict.IGNORE, OnConflict.UPDATE] = None,
            conflict_target: Field = None,
            replace_fields: Sequence[Field] = None,
            returning: Union[Field, Sequence[Field]] = None,
    ):
        """Insert an multiple entities at once into the data store."""
        raise NotImplementedError()

    async def update_all(
            self,
            schema: Type[E],
            values: Mapping,
            where: ExprLike = None,
            returning: Union[Field, Sequence[Field]] = None,
    ) -> List[Mapping]:
        """Update all entities matched by `filters` clause."""
        raise NotImplementedError()

    async def delete_all(
            self, schema: Type[E], where: ExprLike = None,
    ) -> None:
        """Delete all entities matched by `filters` clause."""
        raise NotImplementedError()

    async def fetch_all(
            self,
            query: Any,
            values: Mapping = None,
    ) -> List[Mapping]:
        raise NotImplementedError()

    def iterate(
            self, query: Any, values: Mapping = None,
    ) -> AsyncGenerator[Mapping, None]:
        raise NotImplementedError()

    async def execute(
            self,
            query: Any,
            values: Union[Mapping, Sequence[Mapping]] = None
    ):
        raise NotImplementedError()

    async def exists(self, query: Any) -> bool:
        raise NotImplementedError()

    async def count(self, query: Any) -> int:
        raise NotImplementedError()

    def to_string(self, query: QueryBuilder) -> str:
        raise NotImplementedError()



from __future__ import annotations

import asyncio
from functools import reduce
from typing import (
    Any,
    Callable,
    Generic,
    List,
    Mapping,
    TYPE_CHECKING,
    Type,
    TypeVar,
    Union)

import sqlalchemy as sa

from aerie.sql import functions as func
from aerie.collections import Collection
from aerie.sql.connections import Connection
from aerie.expressions import ExprLike
from aerie.fields import Field
from aerie.preloading import Preload
from aerie.protocols import Adapter
from aerie.queries.base import BaseQuery
from aerie.query import QueryBuilder
from aerie.row import Row

if TYPE_CHECKING:
    from aerie.schemas import Schema

SQLLike = Union[str, sa.sql.ClauseElement]
And = sa.sql.and_
Or = lambda *args: sa.sql.and_(sa.sql.or_(*args))
Not = sa.sql.not_

RE = TypeVar("RE")  # entity for RawQuery. does not need to be a Schema
E = TypeVar("E", bound="Schema")
S = TypeVar("S", bound="Schema")
Factory = Callable[[Mapping, Type[E]], E]


def _format_sql(sql: str) -> str:
    try:
        import pygments
        import pygments.lexers
        import pygments.formatters

        lexer = pygments.lexers.get_lexer_by_name("sql")
        formatter = pygments.formatters.get_formatter_by_name("console")
        sql = pygments.highlight(sql, lexer, formatter)
    except ImportError:
        pass
    return sql


class DeleteEntityQuery(Generic[E], BaseQuery):
    def __init__(self, adapter: Adapter, entity: E):
        self._adapter = adapter
        self._entity = entity

    async def execute(self) -> None:
        await self._adapter.delete(self._entity)


class UpdateQuery(BaseQuery):
    def __init__(
            self,
            adapter: Adapter,
            schema: Type[Schema],
            values: Mapping,
            where: ExprLike = None,
    ):
        self._schema = schema
        self._adapter = adapter
        self._values = values

        if isinstance(where, list):
            where = reduce(sa.and_, where) if len(where) > 0 else None
        self._where = where

    async def execute(self) -> None:
        await self._adapter.update_all(
            self._schema, self._values, self._where,
        )


class DeleteQuery(BaseQuery):
    def __init__(
            self,
            adapter: Adapter,
            schema: Type[Schema],
            where: ExprLike = None,
    ):
        self._schema = schema
        self._adapter = adapter
        self._where = where

    async def execute(self) -> None:
        await self._adapter.delete_all(self._schema, self._where)


def dumps(fn):
    def wrapper(self: QueryBuilder, *args, **kwargs):
        if self._dump:
            self._dump_writer.write(
                '\n' + _format_sql(str(self))
            )
        return fn(self, *args, **kwargs)

    return wrapper


class SelectQuery(Generic[E], BaseQuery):
    def __init__(
            self,
            adapter: Adapter,
            schema: Type[Schema],
            query: QueryBuilder,
            *,
            preload_one: List[Preload],
            preload_many: List[Preload],
    ):
        self._connection = connection
        self._schema = schema
        self._query = query
        self._preload_one = preload_one
        self._preload_many = preload_many

    async def execute(self) -> Collection[E]:
        entities = [
            self._schema.from_row(
                row, [r._relation for r in self._preload_one]
            )
            for row in await self._connection.fetch_all(
                self._query.get_expr(),
            )
        ]

        await asyncio.gather(*[
            preloader.process_rows(entities)
            for preloader in self._preload_many
        ])

        return Collection(entities)


class SelectColumnQuery(BaseQuery):
    def __init__(self, connection: Connection, query: QueryBuilder,
                 field: Field):
        self._connection = connection
        self._query = query
        self._field = field

    async def execute(self) -> Collection[Any]:
        rows = await self._connection.fetch_all(self._build_expr())
        return Collection([row[self._field.column_name] for row in rows])


class AggregateQuery(BaseQuery):
    def __init__(self, connection: Connection, query: QueryBuilder,
                 **fn: func.Function):
        self._connection = connection
        self._query = query
        self._fns = fn

    async def execute(self) -> Row:
        row = await self._connection.fetch_one(self._build_expr())
        return Row(**{
            col_name: row[col_name]
            for col_name in self._fns.keys()
        })

    async def get_value(self, name: Union[str, Field]) -> Any:
        row = await self
        return getattr(row, str(name))


class CountQuery(BaseQuery):
    def __init__(self, connection: Connection, query: QueryBuilder):
        self._connection = connection
        self._query = query

    async def execute(self) -> int:
        expr = self._build_expr()
        return await self._connection.fetch_val(expr)


class ExistsQuery(BaseQuery):
    def __init__(self, adapter: Connection, query: QueryBuilder,
                 negate: bool = False):
        self._adapter = adapter
        self._query = query
        self._negate = negate

    async def execute(self) -> bool:
        expr = self._build_expr()
        exists = await self._connection.fetch_val(expr)
        return exists is False if self._negate else exists is True



from __future__ import annotations

import enum
from typing import (Any, Dict, Generic, List, Mapping, TYPE_CHECKING, Type,
                    TypeVar, cast)

from .exceptions import (AssociationNotLoaded, DoesNotExist,
                         ImproperlyConfigured, MultipleResults)
from .fields import Field, Fields, PrimaryKey, Relation
from .utils import make_table_name

if TYPE_CHECKING:
    from .protocols import Adapter

E = TypeVar("E")


class State(enum.Enum):
    NEW = 'new'  # new object, not persisted
    LOADED = 'loaded'  # object was loaded from the store
    DELETED = 'deleted'  # object has been deleted


class Metadata:
    def __init__(
            self,
            *,
            schema: Type[Schema],
            source: str,
            fields: Dict[str, Field],
            relations: Dict[str, Relation],
            primary_key: PrimaryKey,
            abstract: bool = False,
            adapter: Adapter = None,
    ):
        self.schema = schema
        self.fields = Fields(fields)
        self.relations = relations
        self.primary_key = primary_key
        self.adapter = adapter
        self.abstract = abstract
        self.source = source


class EntityMetadata:
    state: State = State.NEW
    adapter: Adapter = None


class SchemaBase(type):
    def __new__(mcs, name: str, bases, attrs: dict):
        meta_options = {}
        meta = attrs.pop("Meta", None)
        if getattr(meta, "abstract", False) is True:
            return super().__new__(mcs, name, bases, attrs)

        if meta:
            for key, value in meta.__dict__.items():
                if not key.startswith("_"):
                    meta_options[key] = value

        pk = getattr(meta, "primary_key", None)
        if pk is None:
            pk = "id"  # default pk name is ID
            if pk not in attrs or not isinstance(attrs[pk], Field):
                raise ImproperlyConfigured(
                    f"Could not infer primary key for schema {name}. "
                    f"Define a {name}.Meta.primary_key attribute "
                    'or add a field with name "id".'
                )
            attrs['id'].auto_increment = True

        if isinstance(pk, str):
            pk = [pk]

        fields: Dict[str, Field] = {}
        relations: Dict[str, Relation] = {}
        for key, value in attrs.items():
            if isinstance(value, Field):
                value.attr_name = key
                if value.source_name is None:
                    value.source_name = key
                fields[key] = value
            if isinstance(value, Relation):
                value.attr_name = key
                relations[key] = value

        for relation in relations.values():
            relation.alter_schema_fields(fields)

        primary_key = PrimaryKey(
            *[field for field_name, field in fields.items()
              if field_name in pk]
        )

        # remove duplicate or unnecessary fields from meta options
        meta_options.pop("primary_key", None)  # replaced by PrimaryKey
        meta_options.pop("table_name", None)  # replaced by sa.Table:name

        source = getattr(meta, "table_name", make_table_name(name))

        new_class = super().__new__(mcs, name, bases, attrs)
        new_class.__meta__ = Metadata(
            schema=cast(Type[Schema], new_class),
            fields=fields,
            relations=relations,
            primary_key=primary_key,
            source=source,
            **meta_options,
        )
        new_class.pk = primary_key
        new_class.DoesNotExist = type(
            f"{name}.DoesNotExist", (DoesNotExist,), {}
        )
        new_class.MultipleResults = type(
            f"{name}.MultipleResults", (MultipleResults,), {}
        )
        new_class.AssociationNotLoaded = type(
            f"{name}.AssociationNotLoaded", (AssociationNotLoaded,), {}
        )

        return new_class


class Schema(Generic[E], metaclass=SchemaBase):
    __meta__: Metadata
    pk: PrimaryKey
    DoesNotExist: Type[DoesNotExist]
    MultipleResults: Type[MultipleResults]
    AssociationNotLoaded: Type[AssociationNotLoaded]

    class Meta:
        abstract = True

    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls, *args, **kwargs)
        instance.__meta__ = EntityMetadata()
        return instance

    def __init__(self, **kwargs):
        super().__init__()
        for name, value in kwargs.items():
            setattr(self, name, value)

    def __repr__(self):
        pks = ', '.join([
            f'{f.name}={getattr(self, f.name)}'
            for f in type(self).pk.fields
        ])
        return f'<{self.__class__.__name__}: {pks}>'

    @classmethod
    def from_row(cls, row: Mapping, relations: List[Relation] = None) -> E:
        entity = {}
        relations = relations or []

        for attr_name, field in cls.__meta__.fields.items():
            entity[attr_name] = field.to_python_value(row[field.source_name])

        # create one to one entities
        for relation in relations:
            relation_value = None

            # if all PK fields has None then the entity probably does not exist
            # in the database
            values = [
                True for pk in relation.schema.pk
                if row[pk.column] is not None
            ]
            if len(values) > 0:
                relation_value = relation.schema.from_row({
                    rf.source_name: rf.to_python_value(row[rf.source_name])
                    for rf in relation.schema.__meta__.fields.values()
                })
            entity[relation.attr_name] = relation_value

        return cls(**entity)

    @classmethod
    def deconstruct(cls, entity: Schema) -> Dict[str, Any]:
        return {
            field.source_name: field.to_db_value(
                getattr(entity, field.attr_name, field.default)
            ) for field in cls.__meta__.fields
        }



import re


def chunked(items, size: int):
    result = []
    for value in items:
        result.append(value)
        if len(result) == size:
            yield result
            result = []

    if len(result):
        yield result


def make_table_name(name: str) -> str:
    name = name.strip("_")
    if name.endswith("x"):
        name += "es"

    if not name.endswith("s"):
        name += "s"

    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()


def format_sql(sql: str) -> str:
    try:
        import pygments
        import pygments.lexers
        import pygments.formatters

        lexer = pygments.lexers.get_lexer_by_name("sql")
        formatter = pygments.formatters.get_formatter_by_name("console")
        sql = pygments.highlight(sql, lexer, formatter)
    except ImportError:
        pass
    return sql



from __future__ import annotations

import enum
import functools
from typing import Any, Literal, TYPE_CHECKING, Union

if TYPE_CHECKING:
    from .fields import Field


class Conditions(enum.Enum):
    CONCAT = 'concat'
    LIKE = 'like'
    ILIKE = 'ilike'
    IN = 'in'
    IS = 'is'
    STARTS_WITH = 'starts_with'
    ENDS_WITH = 'ends_with'
    CONTAINS = 'contains'
    MATCH = 'match'
    BETWEEN = 'between'
    EQUALS = 'equals'
    GREATER = 'greater'
    GREATER_OR_EQUAL = 'greater_or_equal'
    LESSER = 'lesser'
    LESSER_OR_EQUAL = 'lesser_or_equal'
    NEGATE = 'negate'
    ADD = 'add'
    RIGHT_ADD = 'right_add'
    SUBTRACT = 'subtract'
    RIGHT_SUBTRACT = 'right_subtract'
    MULTIPLY = 'multiply'
    RIGHT_MULTIPLY = 'right_multiply'
    DIVIDE = 'divide'
    RIGHT_DIVIDE = 'right_divide'
    MODULO = 'modulo'
    RIGHT_MODULO = 'right_modulo'

    # logical
    OR = 'or'
    AND = 'and'


class Expr:
    def __init__(self, field: Field, condition: Conditions, value: Any = None,
                 negate: bool = False):
        self.field = field
        self.condition = condition
        self.value = value
        self.negate = negate

    def __invert__(self) -> Expr:
        self.negate = not self.negate
        return self

    def __str__(self) -> str:
        return (
            f'<{self.__class__.__name__}: '
            f'field={self.field.__class__.__name__} '
            f'condition={self.condition} '
            f'value={self.value} '
            f'negate={self.negate}>'
        )

    def __or__(self, other: Expr) -> LogicalExpr:
        return LogicalExpr(self, other, Conditions.OR)

    def __and__(self, other: Expr) -> LogicalExpr:
        return LogicalExpr(self, other, Conditions.AND)


class LogicalExpr:
    def __init__(self, left: ExprLike, right: ExprLike, condition: Conditions):
        self.left = left
        self.right = right
        self.condition = condition

    def __or__(self, other: ExprLike) -> LogicalExpr:
        return LogicalExpr(self, other, Conditions.OR)

    def __and__(self, other: ExprLike) -> LogicalExpr:
        return LogicalExpr(self, other, Conditions.AND)

    def __str__(self):
        return f'{self.left} {self.condition} {self.right}'


ExprLike = Union[Expr, LogicalExpr]


class Sort(enum.Enum):
    ASC = 'asc'
    DESC = 'desc'


class Nulls(enum.Enum):
    FIRST = 'first'
    LAST = 'last'
    UNDEFINED = 'undefined'


def and_(*exprs: ExprLike):
    return functools.reduce(
        lambda agg, value: agg & value, exprs[1:], exprs[0]
    )


def or_(*exprs: ExprLike):
    return functools.reduce(
        lambda agg, value: agg | value, exprs[1:], exprs[0]
    )


class OrderBy:

    def __init__(
            self,
            field: Field,
            sort: Literal[Sort.ASC, Sort.DESC],
            nulls: Literal[
                Nulls.LAST, Nulls.FIRST, Nulls.UNDEFINED
            ] = Nulls.UNDEFINED,
    ):
        self.field = field
        self.sort = sort
        self.nulls = nulls

    def nulls_first(self) -> OrderBy:
        self.nulls = Nulls.FIRST
        return self

    def nulls_last(self) -> OrderBy:
        self.nulls = Nulls.LAST
        return self

    def asc(self) -> OrderBy:
        self.sort = Sort.ASC
        return self

    def desc(self) -> OrderBy:
        self.sort = Sort.DESC
        return self



from .collections import Collection
from aerie.sql.connections import Connection
from .schemas import Schema
from .store import Store

__all__ = [
    "Connection",
    "Collection",
    "Schema",
    "Store",
]

__version__ = '0.0.1b2'



from typing import TypeVar

from .fields import Relation
from .schemas import Schema

E = TypeVar('E', bound=Schema)


class Preload:
    def __init__(self, relation: Relation):
        self.relation = relation



from typing import (AsyncGenerator, Dict, List, Literal, Mapping, Sequence,
                    Type, TypeVar, Union)

import sqlalchemy as sa
from asyncpg import UniqueViolationError
from more_itertools import chunked
from sqlalchemy.dialects.postgresql import insert

from aerie.sql.base import SQLAdapter
from aerie.sql.connections import Connection, DatabaseURL
from aerie.sql.table import convert_expression, generate_table
from aerie.exceptions import UniqueViolation
from aerie.expressions import Expr, ExprLike, Nulls, OrderBy, Sort
from aerie.fields import Field
from aerie.sql.functions import Function
from aerie.preloading import Preload
from aerie.protocols import OnConflict
from aerie.query import (CompoundQuery, ExceptionQuery, IntersectQuery, Join,
                         QueryBuilder, UnionQuery)
from aerie.schemas import Schema

E = TypeVar('E', bound=Schema)


class PostgresAdapter(SQLAdapter[E]):

    def __init__(self, dsn: Union[str, DatabaseURL]):
        self._connection = Connection(dsn)
        self._tables: Dict[Type[E], sa.Table] = {}
        self._metadata = sa.MetaData()

    async def insert_all(
            self,
            schema: Type[E],
            values: Sequence[Mapping],
            batch_size: int = None,
            on_conflict: Literal[OnConflict.IGNORE, OnConflict.UPDATE] = None,
            conflict_target: Field = None,
            replace_fields: Sequence[Field] = None,
            returning: Union[Field, Sequence[Field]] = None,
    ) -> Sequence[Mapping]:
        returning = [returning] if isinstance(returning, Field) else returning
        batch_size = batch_size or 999_999_999  # avoid IF statement
        table = self._get_table(schema)

        results = []
        for chunk in chunked(values, batch_size):
            stmt = insert(table).values(chunk)
            if on_conflict == OnConflict.UPDATE:
                if not replace_fields:
                    raise ValueError(
                        '"replace_fields" attribute is required '
                        'when OnConflict.UPDATE used.'
                    )
                stmt = stmt.on_conflict_do_update(
                    index_elements=[table.c[conflict_target.source_name]],
                    set_={
                        f.source_name: getattr(stmt.excluded, f.source_name)
                        for f in replace_fields
                    },
                )
            if on_conflict == OnConflict.IGNORE:
                stmt = stmt.on_conflict_do_nothing(
                    index_elements=[
                        table.c[conflict_target.source_name]
                    ] if conflict_target else None,
                )

            if returning:
                stmt = stmt.returning(*[
                    table.c[f.source_name] for f in returning
                ])

            try:
                result = await self.fetch_all(stmt)
            except UniqueViolationError as ex:
                raise UniqueViolation(
                    ex.message, ex.detail, ex.sqlstate, ex.constraint_name,
                    ex.query,
                )
            else:
                results.extend(result or [])
        return results

    async def update_all(
            self,
            schema: Type[E],
            values: Mapping,
            where: ExprLike = None,
            returning: Union[Field, Sequence[Field]] = None,
    ) -> List[Mapping]:
        """Update all rows in the database matching the `where` rules."""
        change_set = {}
        if returning is None:
            returning = []
        if isinstance(returning, Field):
            returning = [returning]
        for attr, value in values.items():
            schema_field = schema.__meta__.fields[attr]
            source_name = schema_field.source_name

            if isinstance(value, Expr):
                value = self._convert_expr(value)
            else:
                value = schema_field.to_db_value(value)
            change_set[source_name] = value

        table = self._get_table(schema)
        stmt = table.update().values(change_set).returning(*[
            table.c[fname.source_name] for fname in returning
        ])
        if where:
            stmt = stmt.where(self._convert_expr(where))
        return await self._connection.fetch_all(stmt)

    async def delete_all(
            self,
            schema: Type[E],
            where: ExprLike = None,
    ) -> None:
        """Delete all rows in the database matching the `where` rules."""
        table = self._get_table(schema)
        stmt = table.delete()
        if where:
            stmt = stmt.where(self._convert_expr(where))
        await self._connection.execute(stmt)

    async def fetch_all(
            self, query: Union[str, QueryBuilder], values: Mapping = None
    ) -> List[Mapping]:
        """Fetch all rows matching query."""
        if isinstance(query, QueryBuilder):
            query = self.to_sql(query)
        return [row async for row in self.iterate(query, values)]

    async def iterate(
            self, query: Union[str, QueryBuilder], values: Mapping = None,
    ) -> AsyncGenerator[Mapping, None]:
        """Iterate over query result set,"""
        if isinstance(query, QueryBuilder):
            query = self.to_sql(query)
        async for row in self._connection.iterate(query, values):
            yield row

    async def execute(
            self,
            query: Union[str, QueryBuilder],
            values: Union[Mapping, Sequence[Mapping]] = None
    ):
        if isinstance(query, QueryBuilder):
            query = self.to_sql(query)
        if isinstance(values, Sequence):
            return await self._connection.execute_many(query, list(values))
        return await self._connection.execute(query, values)

    async def exists(self, query: QueryBuilder) -> bool:
        expr = sa.select([
            sa.sql.exists(
                self._to_sa_select(query)
            )
        ])
        return await self._connection.fetch_val(expr)

    async def count(self, query: QueryBuilder) -> int:
        stmt = self._to_sa_select(query).alias('__count')
        return await self._connection.fetch_val(
            sa.func.count().select().select_from(stmt)
        )

    def to_string(self, query: QueryBuilder) -> str:
        return self.to_sql(query)

    def to_sql(self, query: QueryBuilder) -> str:
        stmt = self._to_sa_select(query)
        return str(
            stmt.compile(
                dialect=self._connection.dialect,
                compile_kwargs={"literal_binds": True}
            )
        )

    def _convert_expr(self, expr: Expr) -> sa.sql.ClauseElement:
        return convert_expression(expr, self._metadata)

    def _get_table(self, schema: Type[E]) -> sa.Table:
        if schema not in self._tables:
            self._tables[schema] = generate_table(schema, self._metadata)
        return self._tables[schema]

    def _to_sa_select(self, query: QueryBuilder) -> sa.sql.Select:
        if isinstance(query, CompoundQuery):
            return self._build_compound_query(query)

        from_schema = query._schema
        fields = []
        for field in query._fields:
            if isinstance(field, Field):
                fields.append(field)
            if isinstance(field, Function):
                fields.append(self._build_function(field))

        # fields = [f for f in query._fields if isinstance(f, Field)]
        # fn_fields = [f for f in query._fields if isinstance(f, Function)]
        table = self._get_table(from_schema)

        tables = [table]
        select_from = table
        select_from = self._build_joins(select_from, query._joins)

        # for preload in query._preload:
        #     pass

        stmt: sa.sql.Select = sa.sql.select(tables)
        stmt = stmt.select_from(select_from)
        stmt = stmt.with_only_columns([table.c[f.source_name] for f in fields])
        stmt = stmt.where(sa.sql.and_(*[
            self._convert_expr(expr) for expr in query._where
        ]))
        stmt = stmt.having(sa.sql.and_(*[
            self._convert_expr(expr) for expr in query._having
        ]))
        stmt = stmt.group_by(*[
            table.c[f.source_name] for f in query._group_by
        ])

        self._build_order_by(stmt, query._order_by)

        if query._limit is not None:
            stmt = stmt.limit(query._limit)

        if query._offset is not None:
            stmt = stmt.offset(query._offset)

        return stmt

    def _build_function(self, fn: Function) -> sa.sql.ClauseElement:
        field = fn._field
        table = self._get_table(field.schema)
        column = table.c[field.source_name]
        return fn.__expr__(column)

    def _build_preloads(
            self,
            stmt: sa.sql.ClauseElement,
            preloads: List[Preload],
    ):
        pass

    def _build_joins(
            self,
            select_from: sa.sql.TableClause,
            joins: List[Join],
    ) -> sa.sql.TableClause:
        for join in joins:
            right_table = self._get_table(join._right_schema)
            select_from = select_from.join(
                right_table,
                onclause=self._convert_expr(join._on),
                isouter=join._is_outer,
                full=join._is_full
            )
        return select_from

    def _build_compound_query(self, query: CompoundQuery):
        left_stmt = self._to_sa_select(query._left_qs)
        right_stmt = self._to_sa_select(query._right_qs)

        if isinstance(query, UnionQuery):
            return self._build_union_query(left_stmt, right_stmt, query._all)

        if isinstance(query, ExceptionQuery):
            return self._build_except_query(left_stmt, right_stmt, query._all)

        if isinstance(query, IntersectQuery):
            return self._build_intersect_query(left_stmt, right_stmt,
                                               query._all)

    def _build_union_query(
            self,
            left_stmt: sa.sql.ClauseElement,
            right_stmt: sa.sql.ClauseElement,
            all_: bool,
    ):
        fn = sa.sql.union_all if all_ else sa.sql.union
        return fn(left_stmt, right_stmt)

    def _build_except_query(
            self,
            left_stmt: sa.sql.ClauseElement,
            right_stmt: sa.sql.ClauseElement,
            all_: bool,
    ):
        fn = sa.sql.except_all if all_ else sa.sql.except_
        return fn(left_stmt, right_stmt)

    def _build_intersect_query(
            self,
            left_stmt: sa.sql.ClauseElement,
            right_stmt: sa.sql.ClauseElement,
            all_: bool,
    ):
        fn = sa.sql.intersect_all if all_ else sa.sql.intersect
        return fn(left_stmt, right_stmt)

    def _build_order_by(
            self, stmt: sa.sql.ClauseElement,
            order_bys: List[OrderBy],
    ) -> sa.sql.ClauseElement:
        for order_by in order_bys:
            table = self._get_table(order_by.field.schema)
            column = table.c[order_by.field.source_name]
            expr = column.asc()
            if order_by.sort == Sort.DESC:
                expr = column.desc()

            if order_by.nulls == Nulls.FIRST:
                expr = expr.nullsfirst()
            if order_by.nulls == Nulls.LAST:
                expr = expr.nullslast()

            stmt.append_order_by(expr)
        return stmt

    async def __aenter__(self):
        await self._connection.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._connection.disconnect()



from typing import Generic, Mapping, TypeVar

from aerie.exceptions import DeletedEntity
from aerie.protocols import Adapter
from aerie.queries.base import AwaitableQuery
from aerie.schemas import Schema, State

T = TypeVar('T', bound=Schema)


class UpdateObjectQuery(Generic[T], AwaitableQuery):
    def __init__(self, adapter: Adapter, entity: T, values: Mapping):
        self._adapter = adapter
        self._entity = entity
        self._values = values

    async def execute(self) -> T:
        schema = type(self._entity)

        # quick checks
        if self._entity.__meta__.state == State.DELETED:
            raise DeletedEntity(
                'Entity (pk=%s, schema=%s) has been previously deleted '
                'thus cannot be updated.' % (self._entity.pk, schema.__name__),
                self._entity,
            )

        result = await self._adapter.update_all(
            schema=schema,
            values=self._values,
            where=schema.pk == self._entity.pk,
            returning=[
                schema.__meta__.fields[name]
                for name in self._values.keys()
            ]
        )

        for column, value in result[0].items():
            field = schema.__meta__.fields.by_source_name(column)
            setattr(self._entity, field.attr_name, value)

        return self._entity



from __future__ import annotations

from typing import Any

from aerie.protocols import Adapter


class AwaitableQuery:
    async def execute(self) -> Any:
        raise NotImplementedError()

    def __await__(self):
        return self.execute().__await__()


class BaseQuery(AwaitableQuery):
    _adapter: Adapter



from __future__ import annotations

from copy import copy
from typing import Any, Callable, List, Sequence, Type, Union

from .expressions import Expr, OrderBy
from .fields import Field, Relation
from .preloading import Preload
from .schemas import Schema


class Join:
    INNER = 'inner'
    LEFT = 'left'
    RIGHT = 'right'
    FULL = 'full'

    def __init__(
            self,
            left_schema: Type[Schema],
            right_schema: Type[Schema],
            on: Expr = None,
            outer: bool = False,
            full: bool = False,
    ):
        self._left_schema = left_schema
        self._right_schema = right_schema
        # self._type = type
        self._on = on
        self._is_outer = outer
        self._is_full = full


class CompoundQuery:
    def __init__(
            self,
            left_qs: QueryBuilder,
            right_qs: QueryBuilder,
            order_by: List[Field] = None,
            group_by: List[Field] = None,
            limit: int = None,
            offset: int = None,
            all_: bool = None,
    ):
        self._left_qs = left_qs
        self._right_qs = right_qs
        self._order_by = order_by or []
        self._group_by = group_by or []
        self._limit = limit
        self._offset = offset
        self._all = all_

    def order_by(self, *fields: Field) -> CompoundQuery:
        if not len(fields):
            self._order_by = []
        else:
            self._order_by = fields
        return self.clone()

    def group_by(self, *fields: Field) -> CompoundQuery:
        if not len(fields):
            self._group_by = []
        else:
            self._group_by = fields
        return self.clone()

    def limit(self, size: int) -> CompoundQuery:
        self._limit = size
        return self.clone()

    def offset(self, offset: int) -> CompoundQuery:
        self._offset = offset
        return self.clone()

    def page(self, page: int = 1, page_size: int = 50) -> CompoundQuery:
        page = max(page - 1, 0)
        start_offset = page * page_size
        return self.limit(page_size).offset(start_offset)

    def clone(self) -> CompoundQuery:
        return CompoundQuery(
            left_qs=self._left_qs,
            right_qs=self._right_qs,
            order_by=list(self._order_by).copy(),
            group_by=list(self._group_by).copy(),
            limit=self._limit,
            offset=self._offset,
            all_=self._all,
        )


class UnionQuery(CompoundQuery):
    pass


class IntersectQuery(CompoundQuery):
    pass


class ExceptionQuery(CompoundQuery):
    pass


class QueryBuilder:
    def __init__(
            self,
            schema: Type[Schema],
            distinct: bool = None,
            distinct_on: List[Field] = None,
            fields: Sequence[Field] = None,
            where: List[Expr] = None,
            having: List[Expr] = None,
            order_by: List[OrderBy] = None,
            group_by: List[Field] = None,
            preload: List[Preload] = None,
            joins: List[Join] = None,
            limit: int = None,
            offset: int = None,
    ):
        self._schema = schema
        self._fields = fields or schema.__meta__.fields.values()
        self._joins = joins or []
        self._distinct = distinct
        self._distinct_on = distinct_on
        self._where = where or []
        self._having = having or []
        self._order_by = order_by or []
        self._group_by = group_by or []
        self._limit = limit
        self._offset = offset
        self._preload = preload

    def clone(self, **kwargs) -> QueryBuilder:
        current = dict(
            schema=self._schema,
            fields=copy(self._fields),
            where=copy(self._where),
            having=copy(self._having),
            order_by=copy(self._order_by),
            group_by=copy(self._group_by),
            preload=copy(self._preload),
            joins=copy(self._joins),
            limit=self._limit,
            offset=self._offset,
            distinct=self._distinct,
            distinct_on=self._distinct_on,
        )
        current.update(kwargs)

        return self.__class__(**current)

    def select(self, *fields: Field) -> QueryBuilder:
        """Set columns to select."""
        return self.clone(fields=list(fields))

    def add_select(self, field: Field) -> QueryBuilder:
        """Add a column to select."""
        return self.clone(
            fields=[*self._fields, field]
        )

    def filter(self, **kwargs) -> QueryBuilder:
        query = self
        for k, v in kwargs.items():
            field = self._schema.__meta__.fields[k]
            query = self.where(field == v)
        return query

    def where(self, *criteria: Expr) -> QueryBuilder:
        if not len(criteria):
            self._where = []
        else:
            self._where.extend(criteria)
        return self.clone()

    def where_when(
            self,
            condition: Union[Callable[[Any], bool], bool],
            *criteria,
    ) -> QueryBuilder:
        """Add where statement only when condition is truthy."""
        condition = condition() if callable(condition) else condition
        if condition:
            return self.where(*criteria)
        return self.clone()

    def exclude(self, *criteria: Expr) -> QueryBuilder:
        """Exclude results matching criteria from the result set."""
        return self.where(*map(lambda x: ~x, criteria))

    def having(self, *criteria: Expr) -> QueryBuilder:
        """Add HAVING """
        if not len(criteria):
            self._having = []
        else:
            self._having.extend(criteria)
        return self.clone()

    def limit(self, size: int) -> QueryBuilder:
        self._limit = size
        return self.clone()

    def offset(self, offset: int) -> QueryBuilder:
        self._offset = offset
        return self.clone()

    def page(self, page: int = 1, page_size: int = 20) -> QueryBuilder:
        page = max(page - 1, 0)
        start_offset = page * page_size
        return self.limit(page_size).offset(start_offset)

    def order_by(self, *rules: OrderBy) -> QueryBuilder:
        if not len(rules):
            self._order_by = []
        else:
            self._order_by = rules
        return self.clone()

    def group_by(self, *fields: Field) -> QueryBuilder:
        if not len(fields):
            self._group_by = []
        else:
            self._group_by = fields
        return self.clone()

    def hide(self, *fields: Field):
        """Remove fields from the result rows."""
        return self.clone(
            fields=[
                f for f in self._fields
                if hash(f) not in [hash(f2) for f2 in fields]
            ],
        )

    def join(
            self,
            schema: Type[Schema],
            on: Expr = None,
            outer: bool = False,
            full: bool = False,
    ) -> QueryBuilder:
        self._joins.append(
            Join(self._schema, schema, on, outer, full)
        )
        return self.clone()

    def left_join(self, schema: Type[Schema], on: Expr) -> QueryBuilder:
        return self.join(schema, on, True)

    def preload(self, *relations: Relation) -> QueryBuilder:
        for relation in relations:
            self._preload.append(Preload(relation))
        return self.clone()

    def union(self, qs: QueryBuilder, all_: bool = False) -> CompoundQuery:
        return UnionQuery(self, qs, all_=all_)

    def union_all(self, qs: QueryBuilder) -> CompoundQuery:
        return self.union(qs, all_=True)

    def intersection(self, qs: QueryBuilder,
                     all_: bool = False) -> CompoundQuery:
        """Return the rows that are common to all queries."""
        return IntersectQuery(self, qs, all_=all_)

    def intersection_all(self, qs: QueryBuilder) -> CompoundQuery:
        """Return the rows that are common to all the queries."""
        return self.intersection(qs, all_=True)

    def difference(self, qs: QueryBuilder,
                   all_: bool = False) -> CompoundQuery:
        """List the rows in the first that are not in the second."""
        return ExceptionQuery(self, qs, all_=all_)

    def difference_all(self, qs: QueryBuilder) -> CompoundQuery:
        """List the rows in the first that are not in the second."""
        return self.difference(qs, all_=True)



class Row:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def __data__(self):
        return {
            k: v if not isinstance(v, Row) else v.__data__()
            for k, v in self.__dict__.items()
            if not k.startswith("_")
        }

    def __str__(self):
        cols_limit = 5
        data = ", ".join(
            [
                f"{k}={v}"
                for k, v in list(self.__dict__.items())[:cols_limit]
                if not k.startswith("_") or not k.isupper()
            ]
        )
        total_keys = len(self.__dict__.keys())
        if total_keys > cols_limit:
            data += f" and {total_keys - cols_limit} columns more"
        return f"<Row: {data}>"



from __future__ import annotations

from typing import (Any, Callable, Generator, Generic, List, Mapping, Optional,
                    Sequence, TYPE_CHECKING, TypeVar, Union)

from aerie.utils import chunked

if TYPE_CHECKING:
    from aerie.fields import Field

E = TypeVar("E")


def attribute_reader(obj, attr, default=None) -> Any:
    return getattr(obj, attr, default)


class Collection(Generic[E]):
    def __init__(self, items: Sequence[Any]):
        self.items = items
        self._position = 0

    def first(self) -> Optional[E]:
        try:
            return next(self)
        except StopIteration:
            return None

    def last(self) -> Optional[E]:
        ...

    def chunk(self, batch: int) -> Generator[E, None, None]:
        return chunked(self, batch)

    def pluck(self, key: Union[str, Field]) -> Collection[Any]:
        key = str(key)
        return Collection([attribute_reader(item, key, None) for item in self])

    def all(self) -> Sequence[E]:
        return self.items

    def avg(self, field: Union[str, Field]) -> float:
        ...

    def min(self, field: Union[str, Field]) -> float:
        ...

    def mean(self, field: Union[str, Field]) -> float:
        ...

    def max(self, field: Union[str, Field]) -> float:
        ...

    def sum(self, field: Union[str, Field]) -> float:
        ...

    def mode(self, field: Union[str, Field]) -> float:
        # https://en.wikipedia.org/wiki/Mode_(statistics)
        ...

    def map(self, fn: Callable[[E, int], Any]):
        return type(self)(map(fn, list(self)))

    def each(self, fn: Callable[[E, int], None]) -> None:
        ...

    def every(self, fn: Callable[[E, int], bool]) -> bool:
        """Test if all items match the condition specified by `fn`."""
        return all(map(fn, self))

    def pop(self) -> E:
        ...

    def prepend(self, item: E) -> Collection[E]:
        ...

    def append(self, item: E) -> Collection[E]:
        ...

    def except_(self, fn: Callable[[E, int], bool]) -> Collection[E]:
        ...

    def filter(self, fn: Callable[[E, int], bool]) -> Collection[E]:
        ...

    def reduce(self, fn: Callable[[E, int], bool]) -> Collection[E]:
        ...

    def sort(self) -> Collection[E]:
        ...

    def sort_by(self, key, dir="asc") -> Collection[E]:
        ...

    def group_by(self):
        ...

    def as_list(self) -> List[E]:
        return list(self)

    def __getitem__(self, index: Union[int, slice]) -> E:
        if isinstance(index, slice):
            return [
                self._process_item(item)
                for item in self.items[index.start: index.stop: index.step]
            ]
        return self._process_item(self.items[index])

    def __setitem__(self, key, value):
        raise NotImplementedError()

    def __delitem__(self, key):
        raise NotImplementedError()

    def __len__(self) -> int:
        return len(self.items)

    def __iter__(self) -> Collection[E]:
        return self

    def __next__(self) -> E:
        if self._position < len(self):
            entity = self._process_item(self.items[self._position])
            self._position += 1
            return entity
        raise StopIteration()

    def __contains__(self, item: E) -> bool:
        return item in self.items

    def __str__(self) -> str:
        truncate = 10
        remainder = 0
        if len(self) > truncate:
            remainder = len(self) - truncate

        contents = ",".join(map(str, self[0:10]))
        suffix = ""
        if remainder:
            suffix = f" and {remainder} items more"

        return f"[{contents}{suffix}]"

    def _process_item(self, item: Mapping) -> E:
        return item



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aerie', 'aerie.queries', 'aerie.sql']

package_data = \
{'': ['*']}

install_requires = \
['asyncpg>=0.18.3,<0.19.0',
 'databases>=0.2.5,<0.3.0',
 'more_itertools>=8.2.0,<9.0.0',
 'psycopg2-binary>=2.8.4,<3.0.0']

setup_kwargs = {
    'name': 'aerie',
    'version': '0.0.1b2',
    'description': 'An object mapper for async python.',
    'long_description': '<p align="center">\n<a href="https://travis-ci.org/alex-oleshkevich/aerie">\n    <img src="https://api.travis-ci.com/alex-oleshkevich/aerie.svg?branch=master" alt="Build Status">\n</a>\n<a href="https://codecov.io/gh/alex-oleshkevich/aerie">\n    <img src="https://codecov.io/gh/alex-oleshkevich/aerie/branch/master/graph/badge.svg" alt="Coverage">\n</a>\n<a href="https://pypi.org/project/aerie/">\n    <img src="https://badge.fury.io/py/aerie.svg" alt="Package version">\n</a>\n</p>\n\n---\n\n# An object mapper for async python.\n\n## Installation\n\n```bash\npip install aerie\n```\n\n## Usage\n\n```python\nfrom aerie import Connection, Store, Schema, fields\n\nconnection = Connection(\'postgresql://user:pass@localhost/db_name\')\nstore = Store(connection)\n\nclass User(Schema):\n    id = fields.IntegerField()\n    name = fields.String()\n\nrows = await store.raw(\'select * from users\').to(User)\n\n```\n',
    'author': 'alex.oleshkevich',
    'author_email': 'alex.oleshkevich@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/alex-oleshkevich/aerie',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from typing import (Any, AsyncGenerator, Generic, Mapping, Optional,
                    Type, TypeVar)

from aerie.collections import Collection
from aerie.exceptions import DoesNotExist, MultipleResults
from aerie.protocols import Adapter
from aerie.queries.base import AwaitableQuery
from aerie.row import Row

T = TypeVar('T')
MAP_TO = TypeVar('MAP_TO')


class RawQuery(Generic[T], AwaitableQuery):
    entity: Type[T]

    def __init__(
            self,
            adapter: Adapter,
            query: str,
            values: Mapping = None,
            map_to: Type[T] = Row,
    ):
        self._query = query
        self._adapter = adapter
        self._values = values
        self._map_to = map_to

    def map_to(self, map_to: Type[MAP_TO]) -> RawQuery[MAP_TO]:
        return self.clone(map_to=map_to)

    def clone(self, **kwargs: Any) -> RawQuery[T]:
        defaults = dict(
            adapter=self._adapter,
            query=self._query,
            values=self._values,
            map_to=self._map_to,
        )
        defaults.update(kwargs)
        return type(self)(**defaults)

    async def all(self) -> Collection[T]:
        rows = await self._adapter.fetch_all(self._query, self._values)
        return Collection([
            self._map_to(**row) for row in rows
        ])

    async def one(self) -> T:
        rows = Collection([
            self._map_to(**row)
            for row in await self._adapter.fetch_all(self._query, self._values)
        ])
        if len(rows) == 0:
            raise DoesNotExist("Exactly one row expected but zero found.")

        if len(rows) > 1:
            raise MultipleResults(
                f"Exactly one row expected but zero {len(rows)} found."
            )
        return rows.first()

    async def one_or_none(self) -> Optional[T]:
        try:
            return await self.one()
        except DoesNotExist:
            return None

    async def execute(self) -> Any:
        return await self._adapter.execute(self._query, self._values)

    async def iterate(self) -> AsyncGenerator[T, None]:
        async for row in self._adapter.iterate(self._query, self._values):
            yield self._map_to(**row)

    def __str__(self):
        return self._query



from typing import Generic, TypeVar

from aerie.protocols import Adapter
from aerie.queries.base import AwaitableQuery
from aerie.schemas import Schema, State

T = TypeVar('T', bound=Schema)


class InsertObjectQuery(Generic[T], AwaitableQuery):

    def __init__(self, adapter: Adapter, entity: T):
        self._adapter = adapter
        self._entity = entity

    async def execute(self) -> T:
        schema = type(self._entity)
        value = schema.deconstruct(self._entity)
        # remove pk field from value if they are nulls
        for field in schema.pk:
            if value[field.source_name] is None:
                del value[field.source_name]

        result = await self._adapter.insert_all(
            schema=schema,
            values=[value],
            returning=[*schema.pk],
        )

        self._entity.pk = [result[0][field.source_name] for field in schema.pk]
        self._entity.__meta__.state = State.LOADED
        return self._entity



from __future__ import annotations

import sys
from typing import (Any, AsyncGenerator, Generic, Optional, TextIO, Tuple,
                    TypeVar, Union)

from aerie.collections import Collection
from aerie.protocols import Adapter
from aerie.queries.base import AwaitableQuery
from aerie.query import QueryBuilder
from aerie.schemas import Schema
from aerie.utils import format_sql

T = TypeVar('T', bound=Schema)


class SelectQuery(Generic[T], AwaitableQuery):
    def __init__(self, adapter: Adapter, query: QueryBuilder):
        self._adapter = adapter
        self._schema = query._schema
        self._query = query

    async def all(self) -> Collection[T]:
        rows = await self._adapter.fetch_all(self._query)
        return Collection([self._schema.from_row(row) for row in rows])

    execute = all

    async def one(self) -> T:
        self._query = self._query.limit(2)
        rows = await self.all()
        if len(rows) == 0:
            raise self._schema.DoesNotExist(
                "Exactly one row expected but zero found.")

        if len(rows) > 1:
            raise self._schema.MultipleResults(
                f"Exactly one row expected but zero {len(rows)} found."
            )
        return rows.first()

    async def one_or_none(self) -> Optional[T]:
        try:
            return await self.one()
        except self._schema.DoesNotExist:
            return None

    async def iterate(self) -> AsyncGenerator[T, None]:
        async for row in self._adapter.iterate(self._query):
            yield self._schema.from_row(row)

    async def scalar(self) -> Optional[Union[Any, Tuple[Any]]]:
        rows = await self._adapter.fetch_all(self._query)
        if len(rows) == 0:
            return None
        row = rows[0]
        values = tuple([row[f.source_name] for f in self._query._fields])
        if len(values) == 1:
            return values[0]
        return values

    async def exists(self) -> bool:
        return await self._adapter.exists(self._query)

    async def count(self) -> int:
        return await self._adapter.count(self._query)

    def dump(self, writer: TextIO = sys.stdout) -> SelectQuery[T]:
        formatted = format_sql(str(self))
        writer.write(formatted)
        return self

    def __str__(self):
        return self._adapter.to_string(self._query)



from __future__ import annotations

import sqlalchemy as sa
from typing import Any

from aerie.fields import Field


class Function:
    name: str
    func: Any

    def __init__(self, field: Field, *args, alias: str = None):
        self._field = field
        self._args = args
        self._alias = alias or field.attr_name + '_' + self.name

    def alias(self, alias: str) -> Function:
        self._alias = alias
        return self

    def __expr__(self, column: sa.Column) -> sa.sql.ClauseElement:
        return self.func(column).label(self._alias)

    def __str__(self):
        return f'<Function: name={self.name} field={self._field}>'


class Avg(Function):
    func = sa.sql.func


class Min(Function): ...


class Max(Function): ...


class Sum(Function): ...


class Count(Function): ...


class Round(Function): ...


class Abs(Function): ...


class Ceil(Function): ...


class Floor(Function): ...


class MD5(Function): ...


class Upper(Function): ...


class Lower(Function): ...
